/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package eu.agrosense.client.calendar.api.gui;

import com.sun.javafx.css.StyleableBooleanProperty;
import com.sun.javafx.css.StyleableDoubleProperty;
import com.sun.javafx.css.StyleableProperty;
import com.sun.javafx.css.converters.BooleanConverter;
import com.sun.javafx.css.converters.SizeConverter;

import eu.agrosense.client.calendar.api.internal.gui.Bar;
import eu.agrosense.client.calendar.api.model.CalendarNode;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.DoubleProperty;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.ObjectPropertyBase;
import javafx.beans.property.ReadOnlyDoubleProperty;
import javafx.beans.property.ReadOnlyDoubleWrapper;
import javafx.beans.property.SimpleDoubleProperty;
import javafx.beans.value.WritableValue;
import javafx.collections.FXCollections;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;
import javafx.geometry.Dimension2D;
import javafx.geometry.Side;
import javafx.scene.chart.Axis;

/**
 * Code from CategoryAxis, needs to clean up
 * 
 * @author post
 * 
 */
public class CalendarCategoryAxis extends Axis<CalendarNode> {

	private static final int ZERO_VALUE = 20; // where will first bar be located
	private static final int BAR_SHIFT = 30; // vertical bar shift
	private static final int PADDING = Bar.BAR_HEIGHT + BAR_SHIFT; // space
																	// between
																	// series
	private static final double DEFAULT_WIDTH = 100d;

	private int scrollIndex = 0;
	private List<CalendarNode> allDataCategories = new ArrayList<>();
	private final DoubleProperty firstCategoryPos = new SimpleDoubleProperty(
			this, "firstCategoryPos", 0);

	private ListChangeListener<CalendarNode> itemsListener = new ListChangeListener<CalendarNode>() {
		@Override
		public void onChanged(Change<? extends CalendarNode> c) {
			requestAxisLayout();
		}
	};

	/**
	 * The margin between the axis start and the first tick-mark
	 */
	private DoubleProperty startMargin = new StyleableDoubleProperty(5) {
		@Override
		protected void invalidated() {
			requestAxisLayout();
		}

		@Override
		public StyleableProperty<?, ?> getStyleableProperty() {
			return StyleableProperties.START_MARGIN;
		}

		@Override
		public Object getBean() {
			return CalendarCategoryAxis.this;
		}

		@Override
		public String getName() {
			return "startMargin";
		}
	};

	public final double getStartMargin() {
		return startMargin.getValue();
	}

	public final void setStartMargin(double value) {
		startMargin.setValue(value);
	}

	public final DoubleProperty startMarginProperty() {
		return startMargin;
	}

	/**
	 * The margin between the last tick mark and the axis end
	 */
	private DoubleProperty endMargin = new StyleableDoubleProperty(5) {
		@Override
		protected void invalidated() {
			requestAxisLayout();
		}

		@Override
		public StyleableProperty<?, ?> getStyleableProperty() {
			return StyleableProperties.END_MARGIN;
		}

		@Override
		public Object getBean() {
			return CalendarCategoryAxis.this;
		}

		@Override
		public String getName() {
			return "endMargin";
		}
	};

	public final double getEndMargin() {
		return endMargin.getValue();
	}

	public final void setEndMargin(double value) {
		endMargin.setValue(value);
	}

	public final DoubleProperty endMarginProperty() {
		return endMargin;
	}

	/**
	 * If this is true then half the space between ticks is left at the start
	 * and end
	 */
	private BooleanProperty gapStartAndEnd = new StyleableBooleanProperty(true) {
		@Override
		protected void invalidated() {
			requestAxisLayout();
		}

		@Override
		public StyleableProperty<?, ?> getStyleableProperty() {
			return StyleableProperties.GAP_START_AND_END;
		}

		@Override
		public Object getBean() {
			return CalendarCategoryAxis.this;
		}

		@Override
		public String getName() {
			return "gapStartAndEnd";
		}
	};

	public final boolean isGapStartAndEnd() {
		return gapStartAndEnd.getValue();
	}

	public final void setGapStartAndEnd(boolean value) {
		gapStartAndEnd.setValue(value);
	}

	public final BooleanProperty gapStartAndEndProperty() {
		return gapStartAndEnd;
	}

	/**
	 * The ordered list of categories plotted on this axis. This is set
	 * automatically based on the charts data if autoRanging is true. Duplicate
	 * categories may cause odd results.
	 */
	private ObjectProperty<ObservableList<CalendarNode>> categories = new ObjectPropertyBase<ObservableList<CalendarNode>>() {
		ObservableList<CalendarNode> old;

		@Override
		protected void invalidated() {
			final ObservableList<CalendarNode> newItems = get();
			if (old != newItems) {
				// Add and remove listeners
				if (old != null) {
					old.removeListener(itemsListener);
				}
				if (newItems != null) {
					newItems.addListener(itemsListener);
				}
				old = newItems;
			}
		}

		@Override
		public Object getBean() {
			return CalendarCategoryAxis.this;
		}

		@Override
		public String getName() {
			return "categories";
		}
	};

	public final void setCategories(ObservableList<CalendarNode> value) {
		categories.set(value);
		requestAxisLayout();
	}

	/**
	 * Returns a {@link ObservableList} of categories plotted on this axis.
	 * 
	 * @return ObservableList of categories for this axis.
	 * @see #categories
	 */
	public final ObservableList<CalendarNode> getCategories() {
		return categories.get();
	}

	/**
	 * This is the gap between one category and the next along this axis
	 */
	private final ReadOnlyDoubleWrapper categorySpacing = new ReadOnlyDoubleWrapper(
			this, "categorySpacing", 1);

	public final double getCategorySpacing() {
		return categorySpacing.get();
	}

	public final ReadOnlyDoubleProperty categorySpacingProperty() {
		return categorySpacing.getReadOnlyProperty();
	}

	/**
	 * Create a auto-ranging category axis
	 */
	public CalendarCategoryAxis() {
	}

	/**
	 * Create a category axis with the given categories. This will not
	 * auto-range but be fixed with the given categories.
	 * 
	 * @param categories
	 *            List of the categories for this axis
	 */
	public CalendarCategoryAxis(ObservableList<CalendarNode> categories) {
		setAnimated(false);
		setCategories(categories);
		allDataCategories.addAll(categories);
	}

	private double calculateNewSpacing(double length,
			List<CalendarNode> categories) {
		final Side side = getSide();
		double newCategorySpacing = 1;
		if (side != null && categories != null) {
			double bVal = (isGapStartAndEnd() ? (categories.size())
					: (categories.size() - 1));
			// RT-14092 flickering : check if bVal is 0
			newCategorySpacing = (bVal == 0) ? 1
					: (length - getStartMargin() - getEndMargin()) / bVal;
		}
		// if autoranging is off setRange is not called so we update
		// categorySpacing
		if (!isAutoRanging()) {
			categorySpacing.set(newCategorySpacing / 2);
		}
		return newCategorySpacing;
	}

	private double calculateNewFirstPos(double length, double catSpacing) {
		final Side side = getSide();
		double newPos = 1;
		if (side != null) {
			double offset = ((isGapStartAndEnd()) ? (catSpacing / 2) : (0));
			if (side.equals(Side.TOP) || side.equals(Side.BOTTOM)) { // HORIZONTAL
				newPos = 0 + getStartMargin() + offset;
			} else { // VERTICAL
				newPos = length - getStartMargin() - offset;
			}
		}
		// if autoranging is off setRange is not called so we update first
		// cateogory pos.
		if (!isAutoRanging()) {
			firstCategoryPos.set(newPos);
		}
		return newPos;
	}

	// -------------- PROTECTED METHODS
	// --------------------------------------------------------------------------------
	/**
	 * Called to get the current axis range.
	 * 
	 * @return A range object that can be passed to setRange() and
	 *         calculateTickValues()
	 */
	@Override
	protected Object getRange() {
		return new Object[] { getCategories(), categorySpacing.get(),
				firstCategoryPos.get(), getTickLabelRotation() };
	}

	/**
	 * Called to set the current axis range to the given range. If isAnimating()
	 * is true then this method should animate the range to the new range.
	 * 
	 * @param range
	 *            A range object returned from autoRange()
	 * @param animate
	 *            If true animate the change in range
	 */
	@Override
	protected void setRange(Object range, boolean animate) {
		Object[] rangeArray = (Object[]) range;
		@SuppressWarnings({ "unchecked" })
		List<CalendarNode> data = (List<CalendarNode>) rangeArray[0];
		double newCategorySpacing = (Double) rangeArray[1];
		double newFirstCategoryPos = (Double) rangeArray[2];
		double tickLabelRotation = (Double) rangeArray[3];
		setTickLabelRotation(tickLabelRotation);
		setCategories(FXCollections.<CalendarNode> observableArrayList(data));
		categorySpacing.set(newCategorySpacing);
		firstCategoryPos.set(newFirstCategoryPos);
	}

	/**
	 * This calculates the categories based on the data provided to
	 * invalidateRange() method. This must not effect the state of the axis,
	 * changing any properties of the axis. Any results of the auto-ranging
	 * should be returned in the range object. This will we passed to setRange()
	 * if it has been decided to adopt this range for this axis.
	 * 
	 * @param length
	 *            The length of the axis in screen coordinates
	 * @return Range information, this is implementation dependent
	 */
	@Override
	protected Object autoRange(double length) {
		// TODO check if we can display all categories
		final double newCategorySpacing = calculateNewSpacing(length,
				allDataCategories);
		final double newFirstPos = calculateNewFirstPos(length,
				newCategorySpacing);
		double tickLabelRotation = 0;
		// double requiredLengthToDisplay = calculateRequiredSize(vertical,
		// tickLabelRotation);
		// if (requiredLengthToDisplay > length) {
		// // change text to vertical
		// tickLabelRotation = 90;
		// }
		return new Object[] { allDataCategories, newCategorySpacing,
				newFirstPos, tickLabelRotation };
	}

	/**
	 * Calculate a list of all the data values for each tick mark in range
	 * 
	 * @param length
	 *            The length of the axis in display units
	 * @return A list of tick marks that fit along the axis if it was the given
	 *         length
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected List<CalendarNode> calculateTickValues(double length, Object range) {
		Object[] rangeArray = (Object[]) range;
		// noinspection unchecked
		return (List<CalendarNode>) rangeArray[0];
	}

	@Override
	protected String getTickMarkLabel(CalendarNode value) {
		return value.getName();
	}

	/**
	 * Measure the size of the label for given tick mark value. This uses the
	 * font that is set for the tick marks
	 * 
	 * @param value
	 *            tick mark value
	 * @param range
	 *            range to use during calculations
	 * @return size of tick mark label for given value
	 */
	@Override
	protected Dimension2D measureTickMarkSize(CalendarNode value, Object range) {
		final Object[] rangeArray = (Object[]) range;
		final double tickLabelRotation = (Double) rangeArray[3];
		return measureTickMarkSize(value, tickLabelRotation);
	}

	@Override
	public void invalidateRange(List<CalendarNode> data) {
		super.invalidateRange(data);
		LinkedHashSet<CalendarNode> categoryItems = new LinkedHashSet<>();
		categoryItems.addAll(data);
		allDataCategories.clear();
		allDataCategories.addAll(categoryItems);
	}

	@Override
	public double getDisplayPosition(CalendarNode value) {
		int root = calculateRootPosition(value) + getZero();
		int depth = calculateDepth(value);

		return root + depth * BAR_SHIFT;

	}

	private int calculateRootPosition(CalendarNode model) {
		int index = getCategories().indexOf(getRoot(model));
		int position = ZERO_VALUE;
		for (int i = 0; i < index; i++) {
			CalendarNode root = getCategories().get(i);
			int level = getChildrenLevel(root, 0);
			position += level * BAR_SHIFT + PADDING;
		}
		return position;
	}

	private int getChildrenLevel(CalendarNode model, int level) {
		int maxLevel = level;
		for (CalendarNode child : model.getChildren()) {
			maxLevel++;
			maxLevel = getChildrenLevel(child, maxLevel);
		}
		return maxLevel;
	}

	private int calculateDepth(CalendarNode model) {
		int siblingsDepth = 0;
		CalendarNode parent = model.getParent();
		if (parent != null) {
			for (CalendarNode child : parent.getChildren()) {
				if (child.equals(model)) {
					break;
				} else {
					siblingsDepth += getChildrenLevel(child, 0) + 1;
				}
			}
		}
		int depth = siblingsDepth;
		while (model.getParent() != null) {
			model = model.getParent();
			depth++;
		}
		return depth;
	}

	private CalendarNode getRoot(CalendarNode model) {
		if (model.getParent() == null) {
			return model;
		} else {
			return getRoot(model.getParent());
		}
	}

	/**
	 * Get the data value for the given display position on this axis. If the
	 * axis is a CategoryAxis this will be the nearest value.
	 * 
	 * @param displayPosition
	 *            A pixel position on this axis
	 * @return the nearest data value to the given pixel position or null if not
	 *         on axis;
	 */
	@Override
	public CalendarNode getValueForDisplay(double displayPosition) {
		if (getSide().equals(Side.TOP) || getSide().equals(Side.BOTTOM)) { // HORIZONTAL
			if (displayPosition < 0 || displayPosition > getHeight()) {
				return null;
			}
			double d = (displayPosition - firstCategoryPos.get())
					/ categorySpacing.get();
			return toRealValue(d);
		} else { // VERTICAL
			if (displayPosition < 0 || displayPosition > getWidth()) {
				return null;
			}
			double d = (displayPosition - firstCategoryPos.get())
					/ (categorySpacing.get() * -1);
			return toRealValue(d);
		}
	}

	/**
	 * Checks if the given value is plottable on this axis
	 * 
	 * @param value
	 *            The value to check if its on axis
	 * @return true if the given value is plottable on this axis
	 */
	@Override
	public boolean isValueOnAxis(CalendarNode value) {
		return getCategories().indexOf("" + value) != -1;
	}

	/**
	 * All axis values must be representable by some numeric value. This gets
	 * the numeric value for a given data value.
	 * 
	 * @param value
	 *            The data value to convert
	 * @return Numeric value for the given data value
	 */
	@Override
	public double toNumericValue(CalendarNode value) {
		return getCategories().indexOf(value);
	}

	/**
	 * All axis values must be representable by some numeric value. This gets
	 * the data value for a given numeric value.
	 * 
	 * @param value
	 *            The numeric value to convert
	 * @return Data value for given numeric value
	 */
	@Override
	public CalendarNode toRealValue(double value) {
		int index = (int) Math.round(value);
		List<CalendarNode> cats = getCategories();
		if (index >= 0 && index < cats.size()) {
			return cats.get(index);
		} else {
			return null;
		}
	}

	/**
	 * Get the display position of the zero line along this axis. As there is no
	 * concept of zero on a CategoryAxis this is always Double.NaN.
	 * 
	 * @return always Double.NaN for CategoryAxis
	 */
	@Override
	public double getZeroPosition() {
		return Double.NaN;
	}

	// -------------- STYLESHEET HANDLING
	// ------------------------------------------------------------------------------
	/**
	 * @treatAsPrivate implementation detail
	 */
	@SuppressWarnings("deprecation")
	private static class StyleableProperties {

		@SuppressWarnings("rawtypes")
		private static final List<StyleableProperty> STYLEABLES;

		private static final StyleableProperty<CalendarCategoryAxis, Number> START_MARGIN = new StyleableProperty<CalendarCategoryAxis, Number>(
				"-fx-start-margin", SizeConverter.getInstance(), 5.0) {
			@Override
			public boolean isSettable(CalendarCategoryAxis n) {
				return n.startMargin == null || !n.startMargin.isBound();
			}

			@Override
			public WritableValue<Number> getWritableValue(CalendarCategoryAxis n) {
				return n.startMarginProperty();
			}
		};
		
		private static final StyleableProperty<CalendarCategoryAxis, Number> END_MARGIN = new StyleableProperty<CalendarCategoryAxis, Number>(
				"-fx-end-margin", SizeConverter.getInstance(), 5.0) {
			@Override
			public boolean isSettable(CalendarCategoryAxis n) {
				return n.endMargin == null || !n.endMargin.isBound();
			}

			@Override
			public WritableValue<Number> getWritableValue(CalendarCategoryAxis n) {
				return n.endMarginProperty();
			}
		};

		private static final StyleableProperty<CalendarCategoryAxis, Boolean> GAP_START_AND_END = new StyleableProperty<CalendarCategoryAxis, Boolean>(
				"-fx-gap-start-and-end", BooleanConverter.getInstance(),
				Boolean.TRUE) {
			@Override
			public boolean isSettable(CalendarCategoryAxis n) {
				return n.gapStartAndEnd == null || !n.gapStartAndEnd.isBound();
			}

			@Override
			public WritableValue<Boolean> getWritableValue(
					CalendarCategoryAxis n) {
				return n.gapStartAndEndProperty();
			}
		};

		static {
			@SuppressWarnings("rawtypes")
			final List<StyleableProperty> styleables = new ArrayList<>(
					Axis.impl_CSS_STYLEABLES());
			Collections.addAll(styleables, START_MARGIN, END_MARGIN,
					GAP_START_AND_END);
			STYLEABLES = Collections.unmodifiableList(styleables);
		}
	}

	/**
	 * @treatAsPrivate implementation detail
	 * @deprecated This is an internal API that is not intended for use and will
	 *             be removed in the next version
	 */
	@SuppressWarnings("rawtypes")
	@Deprecated
	public static List<StyleableProperty> impl_CSS_STYLEABLES() {
		return CalendarCategoryAxis.StyleableProperties.STYLEABLES;
	}

	/**
	 * RT-19263
	 * 
	 * @treatAsPrivate implementation detail
	 * @deprecated This is an experimental API that is not intended for general
	 *             use and is subject to change in future versions
	 */
	@SuppressWarnings("rawtypes")
	@Deprecated
	public List<StyleableProperty> impl_getStyleableProperties() {
		return impl_CSS_STYLEABLES();
	}

	protected int getZero() {
		return ZERO_VALUE - getScrollValue();
	}

	protected int getScrollValue() {
		ObservableList<CalendarNode> cats = getCategories();
		if (scrollIndex > 0 && scrollIndex < cats.size()) {
			int root = calculateRootPosition(cats.get(scrollIndex));
			return root;
		} else {
			return 0;
		}
	}

	public boolean scrollUp() {
		ObservableList<CalendarNode> cats = getCategories();
		if (scrollIndex < cats.size() - 1) {
			scrollIndex++;
			return true;
		} else {
			return false;
		}
	}

	public boolean scrollDown() {
		if (scrollIndex > 0) {
			scrollIndex--;
			return true;
		} else {
			return false;
		}
	}

	@Override
	protected double computePrefWidth(double height) {
		return DEFAULT_WIDTH;
	}
}
