/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package eu.agrosense.client.calendar.api.gui;

import java.util.Stack;

import javafx.application.Platform;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.geometry.Insets;
import javafx.geometry.Orientation;
import javafx.geometry.Pos;
import javafx.geometry.Side;
import javafx.scene.chart.XYChart.Series;
import javafx.scene.control.Button;
import javafx.scene.control.ContextMenu;
import javafx.scene.control.Label;
import javafx.scene.control.ListCell;
import javafx.scene.control.ListView;
import javafx.scene.control.MenuItem;
import javafx.scene.control.RadioButton;
import javafx.scene.control.Toggle;
import javafx.scene.control.ToggleButton;
import javafx.scene.control.ToggleGroup;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.input.MouseEvent;
import javafx.scene.input.ScrollEvent;
import javafx.scene.layout.AnchorPane;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.FlowPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Priority;
import javafx.scene.layout.VBox;
import javafx.util.Callback;

import org.joda.time.DateTime;
import org.joda.time.LocalDate;
import org.joda.time.Period;
import org.openide.util.NbBundle;

import eu.agrosense.client.calendar.api.event.AxisEvent;
import eu.agrosense.client.calendar.api.event.AxisEventHandler;
import eu.agrosense.client.calendar.api.event.BarEvent;
import eu.agrosense.client.calendar.api.event.BarEventHandler;
import eu.agrosense.client.calendar.api.impl.gui.LinearAxisCalculator;
import eu.agrosense.client.calendar.api.impl.gui.NonLinearAxisCalculator;
import eu.agrosense.client.calendar.api.internal.gui.CalendarChart;
import eu.agrosense.client.calendar.api.internal.gui.DefaultListCell;
import eu.agrosense.client.calendar.api.internal.util.DateBounds;
import eu.agrosense.client.calendar.api.internal.util.DateUtil;
import eu.agrosense.client.calendar.api.model.CalendarNode;

@NbBundle.Messages({
		"scroll_up_icon=nl/cloudfarming/client/geoviewer/jxmap/icons/scroll_up10.png",
		"scroll_down_icon=nl/cloudfarming/client/geoviewer/jxmap/icons/scroll_down10.png" })
public class CalendarComponent extends VBox {

	private static final int HIDDEN_LIST_VIEW_WIDTH = 30;
	private static final int HIDE_BUTTON_SIZE = 10;
	private static final int BUTTON_WIDTH = 60;
	private static final int BUTTON_HEIGHT = 30;
	private static final int INSET = 5;
	private static final int LIST_VIEW_WIDTH = 150;
	private LocalDate actualDate = LocalDate.now();
	private ToggleButton buttonYear;
	private ToggleButton buttonMonth;
	private ToggleButton buttonWeek;
	private ToggleButton buttonDay;
	private Button buttonToday;
	private Button buttonPrevious;
	private Button buttonNext;
	private Label actualLabel;
	private CalendarChart chart;
	private CalendarDateAxis xAxis;
	private CalendarCategoryAxis yAxis;
	private Button buttonDown;
	private Button buttonUp;
	private Button buttonBack;
	private boolean listViewHidden = false;
	private Image scrollUpIcon = new Image(Bundle.scroll_up_icon(), true);
	private Image scrollDownIcon = new Image(Bundle.scroll_down_icon(), true);
    
	private ToggleGroup boundsButtonsGroup;
	private ChangeListener<Number> rangeChangeListener;
	private DateBounds lastDateBounds;
	Stack<DateBounds> boundsHistory = new Stack<>();
	private Toggle lastToggleButton;

	private double dragStartPosition = -1;
	private AxisEventHandler<DateTime> dateAxisEventHandler;
	
	private RadioButton linearAxisRadio;
	private RadioButton nonLinearAxisRadio;

	public CalendarComponent() {
		super(1);

		rangeChangeListener = new ChangeListener<Number>() {
			@Override
			public void changed(ObservableValue<? extends Number> observable,
					Number oldValue, Number newValue) {
				Object[] range = (Object[]) xAxis.getRange();
				DateTime minDate = (DateTime) range[0];
				DateTime maxDate = (DateTime) range[1];
				updateActualDateLabel(minDate, maxDate);
			}
		};

		setOnMouseReleased(new EventHandler<MouseEvent>() {

			@Override
			public void handle(MouseEvent mouseEvent) {
				resetDragStartPosition();
			}

		});

		setOnMouseDragged(new EventHandler<MouseEvent>() {

			@Override
			public void handle(MouseEvent event) {
				double shift = 0;

				if (dragStartPosition != -1) {
					shift = (event.getX() - dragStartPosition);
					dragStartPosition = event.getX();

					DateTime val1 = xAxis.getValueForDisplay(50);
					DateTime val2 = xAxis.getValueForDisplay(50 + shift);

					long diff = val1.getMillis() - val2.getMillis();
					Period period = new Period(diff);

					Object[] range = (Object[]) xAxis.getRange();
					DateTime min = (DateTime) range[0];
					DateTime max = (DateTime) range[1];
					
					min = min.plus(period);
					max = max.plus(period);
					actualDate = min.toLocalDate();
					DateBounds ld = new DateBounds(new DateTime(min), new DateTime(max));
					setRange(ld);
					chart.refresh();

				} else {
					dragStartPosition = event.getX();
				}
			}
		});

		initComponents();
	}

	protected void resetDragStartPosition() {
		dragStartPosition = -1;
	}

	private void initComponents() {
		FlowPane buttonPane = new FlowPane();
		buttonPane.setAlignment(Pos.TOP_RIGHT);
		FlowPane infoPane = new FlowPane();

		buttonYear = new ToggleButton(org.openide.util.NbBundle.getMessage(
				CalendarComponent.class, "CalendarComponent.button.year"));
		buttonMonth = new ToggleButton(org.openide.util.NbBundle.getMessage(
				CalendarComponent.class, "CalendarComponent.button.month"));
		buttonWeek = new ToggleButton(org.openide.util.NbBundle.getMessage(
				CalendarComponent.class, "CalendarComponent.button.week"));
		buttonDay = new ToggleButton(org.openide.util.NbBundle.getMessage(
				CalendarComponent.class, "CalendarComponent.button.day"));
		buttonToday = new Button(org.openide.util.NbBundle.getMessage(
				CalendarComponent.class, "CalendarComponent.button.today"));
		buttonPrevious = new Button(org.openide.util.NbBundle.getMessage(
				CalendarComponent.class, "CalendarComponent.button.previous"));
		buttonNext = new Button(org.openide.util.NbBundle.getMessage(
				CalendarComponent.class, "CalendarComponent.button.next"));
		actualLabel = new Label();
		buttonBack = new Button(org.openide.util.NbBundle.getMessage(
				CalendarComponent.class, "CalendarComponent.button.back"));
		buttonBack.setVisible(false);

		buttonYear.setPrefSize(BUTTON_WIDTH, BUTTON_HEIGHT);
		buttonMonth.setPrefSize(BUTTON_WIDTH, BUTTON_HEIGHT);
		buttonWeek.setPrefSize(BUTTON_WIDTH, BUTTON_HEIGHT);
		buttonDay.setPrefSize(BUTTON_WIDTH, BUTTON_HEIGHT);
		buttonToday.setPrefSize(BUTTON_WIDTH, BUTTON_HEIGHT);
		buttonPrevious.setPrefSize(BUTTON_HEIGHT, BUTTON_HEIGHT);
		buttonNext.setPrefSize(BUTTON_HEIGHT, BUTTON_HEIGHT);
		buttonBack.setPrefSize(BUTTON_WIDTH, BUTTON_HEIGHT);

		buttonUp = new Button(org.openide.util.NbBundle.getMessage(
				CalendarComponent.class, "CalendarComponent.button.scrollUp"),
				new ImageView(scrollUpIcon));
		buttonDown = new Button(
				org.openide.util.NbBundle.getMessage(CalendarComponent.class,
						"CalendarComponent.button.scrollDown"), new ImageView(
						scrollDownIcon));

		boundsButtonsGroup = new ToggleGroup();
		buttonYear.setToggleGroup(boundsButtonsGroup);
		buttonMonth.setToggleGroup(boundsButtonsGroup);
		buttonWeek.setToggleGroup(boundsButtonsGroup);
		buttonDay.setToggleGroup(boundsButtonsGroup);
		buttonYear.setSelected(true);

		initChart();

		AnchorPane anchorPane = new AnchorPane();
		anchorPane.getChildren().add(buttonPane);
		anchorPane.getChildren().add(infoPane);
		AnchorPane.setLeftAnchor(infoPane, 0d);
		AnchorPane.setTopAnchor(infoPane, 0d);
		anchorPane.setPadding(new Insets(INSET));
		anchorPane.setMinHeight(BUTTON_HEIGHT + 2 * INSET);
		anchorPane.setPrefHeight(BUTTON_HEIGHT + 2 * INSET);

		AnchorPane.setRightAnchor(buttonPane, 0d);
		AnchorPane.setTopAnchor(buttonPane, 0d);
		buttonPane.toFront();

		anchorPane.getStyleClass().add("button_panel");
		getChildren().add(anchorPane);
		getChildren().add(chart);

		final ListView<Series<DateTime, CalendarNode>> listView = new ListView<Series<DateTime, CalendarNode>>(
				chart.getData());
		listView.setOrientation(Orientation.VERTICAL);

		MenuItem menuItemRemove = new MenuItem(org.openide.util.NbBundle.getMessage(CalendarComponent.class,
						"CalendarComponent.items.listView.removeAction"));
		ContextMenu contextMenu = new ContextMenu(menuItemRemove);
		menuItemRemove.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent t) {
				chart.getData().remove(
						(Series<DateTime, CalendarNode>) listView
								.getSelectionModel().getSelectedItem());
			}
		});

		final HBox hBox = new HBox();
		final BorderPane listViewPane = new BorderPane();

		// TODO listView in container with buttons to hide/resume
		listView.getStyleClass().add("list-view");
		listView.setCellFactory(createCellFactory(contextMenu));
		listViewPane.setPrefWidth(LIST_VIEW_WIDTH);
		listViewPane.setMaxWidth(LIST_VIEW_WIDTH);
		listViewPane.setCenter(listView);

		AnchorPane listViewHeaderPane = new AnchorPane();
		listViewHeaderPane.getStyleClass().add("list-view-header");
		listViewHeaderPane.setPrefHeight(20d);
		listViewPane.setTop(listViewHeaderPane);

		final Label headerLabel = new Label("Node list");
		AnchorPane.setLeftAnchor(headerLabel, 2.0);
		AnchorPane.setBottomAnchor(headerLabel, 2.0);
		listViewHeaderPane.getChildren().add(headerLabel);

		Button buttonHide = new Button(".");
		AnchorPane.setTopAnchor(buttonHide, 2.0);
		AnchorPane.setRightAnchor(buttonHide, 2.0);
		buttonHide.setPrefSize(HIDE_BUTTON_SIZE, HIDE_BUTTON_SIZE);
		listViewHeaderPane.getChildren().add(buttonHide);
		buttonHide.setOnMouseClicked(new EventHandler<MouseEvent>() {
			@Override
			public void handle(MouseEvent t) {
				if (!listViewHidden) {
					listViewPane.setPrefWidth(HIDDEN_LIST_VIEW_WIDTH);
					listViewPane.setMaxWidth(HIDDEN_LIST_VIEW_WIDTH);
					headerLabel.setVisible(false);
					listViewHidden = true;
				} else {
					listViewPane.setPrefWidth(LIST_VIEW_WIDTH);
					listViewPane.setMaxWidth(LIST_VIEW_WIDTH);
					headerLabel.setVisible(true);
					listViewHidden = false;
				}
				hBox.layout();
			}
		});

		AnchorPane panel = new AnchorPane();

		AnchorPane.setTopAnchor(buttonUp, 5.0);
		AnchorPane.setLeftAnchor(buttonUp, 0.0);
		AnchorPane.setRightAnchor(buttonUp, 10.0);

		AnchorPane.setBottomAnchor(buttonDown, 5.0);
		AnchorPane.setLeftAnchor(buttonDown, 0.0);
		AnchorPane.setRightAnchor(buttonDown, 10.0);

		panel.getChildren().add(buttonUp);
		panel.getChildren().add(buttonDown);

		
		VBox chartBox = new VBox();
		
		hBox.getStyleClass().add("main-content-box");
		hBox.getChildren().add(listViewPane);
		hBox.getChildren().add(chartBox);
		hBox.getChildren().add(panel);
		hBox.setFillHeight(true);

		HBox.setHgrow(chartBox, Priority.ALWAYS);
		getChildren().add(hBox);
		VBox.setVgrow(hBox, Priority.ALWAYS);
		VBox.setVgrow(chartBox, Priority.ALWAYS);
		VBox.setVgrow(chart, Priority.ALWAYS);
		
		VBox.setVgrow(panel, Priority.NEVER);
		setFillWidth(true);
		
		chartBox.getChildren().add(chart);
		
		linearAxisRadio = new RadioButton(org.openide.util.NbBundle.getMessage(CalendarComponent.class, "CalendarComponent.axis.linear"));
		linearAxisRadio.setSelected(true);
		nonLinearAxisRadio = new RadioButton(org.openide.util.NbBundle.getMessage(CalendarComponent.class, "CalendarComponent.axis.nonlinear"));
		AnchorPane.setLeftAnchor(linearAxisRadio, 110d);
		AnchorPane.setLeftAnchor(nonLinearAxisRadio, 210d);
		AnchorPane.setBottomAnchor(linearAxisRadio, 5d);
		AnchorPane.setBottomAnchor(nonLinearAxisRadio, 5d);
		
		ToggleGroup axisToggleGroup = new ToggleGroup();
		linearAxisRadio.setToggleGroup(axisToggleGroup);
		nonLinearAxisRadio.setToggleGroup(axisToggleGroup);
		
		AnchorPane radioButtonsPane = new AnchorPane();
		radioButtonsPane.getChildren().add(linearAxisRadio);
		radioButtonsPane.getChildren().add(nonLinearAxisRadio);
		chartBox.getChildren().add(radioButtonsPane);
		

		BorderPane separatorPane = new BorderPane();
		separatorPane.getStyleClass().add("separator-pane");

		infoPane.getChildren().add(buttonToday);

		infoPane.getChildren().add(separatorPane);
		infoPane.getChildren().add(buttonPrevious);
		infoPane.getChildren().add(buttonNext);
		infoPane.getChildren().add(buttonBack);
		infoPane.getChildren().add(actualLabel);

		buttonPane.getChildren().add(buttonYear);
		buttonPane.getChildren().add(buttonMonth);
		buttonPane.getChildren().add(buttonWeek);
		buttonPane.getChildren().add(buttonDay);

		initListeners();
		initStyles();
	}

	Callback<ListView<Series<DateTime, CalendarNode>>, ListCell<Series<DateTime, CalendarNode>>> createCellFactory(final ContextMenu contextMenu) {
		
		return new Callback<ListView<Series<DateTime, CalendarNode>>, ListCell<Series<DateTime, CalendarNode>>>() {
			@Override
			public ListCell<Series<DateTime, CalendarNode>> call(ListView<Series<DateTime, CalendarNode>> listView) {
				ListCell<Series<DateTime, CalendarNode>> cell = new DefaultListCell<>();
				cell.setContextMenu(contextMenu);
				return cell;
			}
		};
		
	}

	private void initStyles() {
		getStyleClass().add("calendar-component");
		buttonYear.getStyleClass().add("big-button");
		buttonMonth.getStyleClass().add("big-button");
		buttonWeek.getStyleClass().add("big-button");
		buttonDay.getStyleClass().add("big-button");
		buttonToday.getStyleClass().add("big-button");
		buttonToday.getStyleClass().add("today-button");
		buttonPrevious.getStyleClass().add("big-button");
		buttonNext.getStyleClass().add("big-button");
		actualLabel.getStyleClass().add("actual-date");
		buttonUp.getStyleClass().add("small-button");
		buttonDown.getStyleClass().add("small-button");
		buttonBack.getStyleClass().add("big-button");

	}

	protected CalendarChart initChart() {
		lastDateBounds = new DateBounds(DateUtil.startOfCurrentYear(), DateUtil.endOfCurrentYear());
		xAxis = new CalendarDateAxis(lastDateBounds.getStart(), lastDateBounds.getEnd());
		xAxis.setSide(Side.TOP);
		xAxis.setLabel(org.openide.util.NbBundle.getMessage(
				CalendarComponent.class, "CalendarComponent.chart.xAxis.label"));
		yAxis = new CalendarCategoryAxis();
		yAxis.setLabel(org.openide.util.NbBundle.getMessage(
				CalendarComponent.class, "CalendarComponent.chart.yAxis.label"));
		updateActualDateLabel(lastDateBounds.getStart(), lastDateBounds.getEnd());
		
		chart = new CalendarChart(xAxis, yAxis);
		chart.getStyleClass().add("calendar-chart");

		xAxis.setOnRangePropertyChangeListener(rangeChangeListener);

		BarEventHandler barEventHandler = new BarEventHandler() {
			@Override
			public void handle(final BarEvent barEvent) {

				Platform.runLater(new Runnable() {
					@Override
					public void run() {
						zoomToBar(barEvent);
					}
				});
			}
		};
		chart.setBarEventHandler(barEventHandler);

		return chart;
	}

	public CalendarChart getChart() {
		return chart;
	}

	private void initListeners() {
		buttonYear.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent actionEvent) {
				if (!buttonYear.isSelected()) {
					buttonYear.setSelected(true);
				} else {
					if (lastToggleButton != null && lastToggleButton != buttonYear) {
						lastToggleButton = buttonYear;
					}
					DateBounds ld = DateUtil.getYearBounds(actualDate);
					setRange(ld);
					rangeFromGraph(false);
				}

			}
		});

		buttonMonth.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent actionEvent) {
				if (!buttonMonth.isSelected()) {
					buttonMonth.setSelected(true);
				} else {
					if (lastToggleButton != null && lastToggleButton != buttonMonth) {
						lastToggleButton = buttonMonth;
					}
					DateBounds ld = DateUtil.getMonthBounds(actualDate);
					setRange(ld);
					rangeFromGraph(false);
				}

			}
		});

		buttonWeek.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent actionEvent) {
				if (!buttonWeek.isSelected()) {
					buttonWeek.setSelected(true);
				} else {
					if (lastToggleButton != null && lastToggleButton != buttonWeek) {
						lastToggleButton = buttonWeek;
					}
					DateBounds ld = DateUtil.getWeekBounds(actualDate);
					setRange(ld);
					rangeFromGraph(false);
				}

			}
		});

		buttonDay.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent actionEvent) {
				if (!buttonDay.isSelected()) {
					buttonDay.setSelected(true);
				} else {
					if (lastToggleButton != null && lastToggleButton != buttonDay) {
						lastToggleButton = buttonDay;
					}
					DateBounds ld = DateUtil.getDayBounds(actualDate);
					setRange(ld);
					rangeFromGraph(false);
				}

			}
		});

		
		buttonToday.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent actionEvent) {
				LocalDate dateToday = LocalDate.now();
				actualDate = dateToday;
				if (buttonDay.isSelected()) {
					DateBounds ld = DateUtil.getDayBounds(actualDate);
					setRange(ld);
				} else if (buttonWeek.isSelected()) {
					DateBounds ld = DateUtil.getWeekBounds(actualDate);
					setRange(ld);
				} else if (buttonMonth.isSelected()) {
					DateBounds ld = DateUtil.getMonthBounds(actualDate);
					setRange(ld);
				} else if (buttonYear.isSelected()) {
					DateBounds ld = DateUtil.getYearBounds(actualDate);
					setRange(ld);
				}

			}
		});

		buttonPrevious.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent actionEvent) {
				if (buttonDay.isSelected()) {
					actualDate = DateUtil.previousDay(actualDate);
					DateBounds ld = DateUtil.getDayBounds(actualDate);
					setRange(ld);
				} else if (buttonWeek.isSelected()) {
					actualDate = DateUtil.previousWeek(actualDate);
					DateBounds ld = DateUtil.getWeekBounds(actualDate);
					setRange(ld);
				} else if (buttonMonth.isSelected()) {
					actualDate = DateUtil.previousMonth(actualDate);
					DateBounds ld = DateUtil.getMonthBounds(actualDate);
					setRange(ld);
				} else if (buttonYear.isSelected()) {
					actualDate = DateUtil.previousYear(actualDate);
					DateBounds ld = DateUtil.getYearBounds(actualDate);
					setRange(ld);
				}

			}
		});

		buttonNext.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent actionEvent) {
				if (buttonDay.isSelected()) {
					actualDate = DateUtil.nextDay(actualDate);
					DateBounds ld = DateUtil.getDayBounds(actualDate);
					setRange(ld);
				} else if (buttonWeek.isSelected()) {
					actualDate = DateUtil.nextWeek(actualDate);
					DateBounds ld = DateUtil.getWeekBounds(actualDate);
					setRange(ld);
				} else if (buttonMonth.isSelected()) {
					actualDate = DateUtil.nextMonth(actualDate);
					DateBounds ld = DateUtil.getMonthBounds(actualDate);
					setRange(ld);
				} else if (buttonYear.isSelected()) {
					actualDate = DateUtil.nextYear(actualDate);
					DateBounds ld = DateUtil.getYearBounds(actualDate);
					setRange(ld);
				}

			}
		});

		buttonUp.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent t) {
				if (yAxis.scrollDown()) {
					chart.refresh();
				}
			}
		});

		buttonDown.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent t) {
				if (yAxis.scrollUp()) {
					chart.refresh();
				}
			}
		});

		chart.setOnScroll(new EventHandler<ScrollEvent>() {
			@Override
			public void handle(ScrollEvent t) {
				if (t.getDeltaY() < 0) {
					if (yAxis.scrollUp()) {
						chart.refresh();
					}
				} else {
					if (yAxis.scrollDown()) {
						chart.refresh();
					}
				}
			}
		});

		buttonBack.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent actionEvent) {
				if (!boundsHistory.isEmpty()) {
					DateBounds bounds = boundsHistory.pop();
					setRange(bounds);
					if (boundsHistory.isEmpty()) {
						rangeFromGraph(false);
					}
				}

			}
		});
		
		dateAxisEventHandler = new AxisEventHandler<DateTime>() {
			
			@Override
			public void handle(AxisEvent<DateTime> event) {
				actualDate = event.getValue().toLocalDate();
				//if day range, do nothing
				if (buttonWeek.isSelected()) {
					buttonDay.setSelected(true);
					DateBounds ld = DateUtil.getDayBounds(actualDate);
					setRange(ld);
				} else if (buttonMonth.isSelected()) {
					buttonDay.setSelected(true);
					DateBounds ld = DateUtil.getDayBounds(actualDate);
					setRange(ld);
				} else if (buttonYear.isSelected()) {
					buttonMonth.setSelected(true);
					DateBounds ld = DateUtil.getMonthBounds(actualDate);
					setRange(ld);
				}
				
			}
			
		};
		
		linearAxisRadio.setOnAction(new EventHandler<ActionEvent>() {
			
			@Override
			public void handle(ActionEvent actionEvent) {
				xAxis.setAxisCalculator(LinearAxisCalculator.INSTANCE);
				chart.refresh();
			}
			
		});
		
		nonLinearAxisRadio.setOnAction(new EventHandler<ActionEvent>() {
			
			@Override
			public void handle(ActionEvent actionEvent) {
				xAxis.setAxisCalculator(NonLinearAxisCalculator.INSTANCE);
				chart.refresh();
			}
			
		});

		xAxis.setOnAxisClickedHandler(dateAxisEventHandler);
	}

	protected void updateActualDateLabel(DateTime minDate, DateTime maxDate) {
		setActualLabel(new DateBounds(minDate, maxDate));
	}

	protected void setRange(DateBounds ld) {
		xAxis.setRange(new DateTime[] { ld.getStart(), ld.getEnd() }, false);
		chart.refresh();
		lastDateBounds = ld;
	}

	private void setActualLabel(DateBounds dateBounds) {
		String text = "";
		
		if (rangeInOneDay(dateBounds)) {
			text = dateBounds.getStart().toString("EEEE, MMM d, yyyy");
		} else if (rangeInOneMonth(dateBounds)) {
			text = dateBounds.getStart().toString("MMM d") + " - " + dateBounds.getEnd().toString("d, yyyy");
		} else if (rangeInOneYear(dateBounds)) {
			text = dateBounds.getStart().toString("MMM d") + " - " + dateBounds.getEnd().toString("MMM d, yyyy");
		} else {
			text = dateBounds.getStart().toString("MMM d, yyyy") + " - " + dateBounds.getEnd().toString("MMM d, yyyy");
		}
		
		actualLabel.setText(text);
	}
	
	private boolean rangeInOneDay(DateBounds dateBounds) {
		boolean inOneYear = rangeInOneYear(dateBounds);
		boolean inOneDay = dateBounds.getStart().getDayOfYear() == dateBounds.getEnd().getDayOfYear();
		return inOneYear && inOneDay;
	}
	
	private boolean rangeInOneMonth(DateBounds dateBounds) {
		boolean inOneYear = rangeInOneYear(dateBounds);
		boolean inOneMonth = dateBounds.getStart().getMonthOfYear() == dateBounds.getEnd().getMonthOfYear();
		return inOneYear && inOneMonth;
	}
	
	private boolean rangeInOneYear(DateBounds dateBounds) {
		return dateBounds.getStart().getYear() == dateBounds.getEnd().getYear();
	}

	void rangeFromGraph(boolean fromGraph) {
		if (fromGraph) {
			if (boundsButtonsGroup.getSelectedToggle() != null) {
				lastToggleButton = boundsButtonsGroup.getSelectedToggle();
				boundsButtonsGroup.selectToggle(null);
			}

			Object[] range = (Object[]) xAxis.getRange();
			DateTime minDate = (DateTime) range[0];
			DateTime maxDate = (DateTime) range[1];
			DateBounds dateBounds = new DateBounds(minDate, maxDate);
			if (lastDateBounds != null) {
				boundsHistory.push(lastDateBounds);
			}
			lastDateBounds = dateBounds;
		} else {
			boundsHistory.clear();

			if (lastToggleButton != null) { 
				if (boundsButtonsGroup.getSelectedToggle() != lastToggleButton) {
					boundsButtonsGroup.selectToggle(lastToggleButton);
				}
				lastToggleButton = null;
			}
		}
		buttonNext.setDisable(fromGraph);
		buttonPrevious.setDisable(fromGraph);
		buttonBack.setVisible(fromGraph);
	}

	private void zoomToBar(BarEvent barEvent) {
		CalendarNode model = barEvent.getNode();
		Object[] range = (Object[]) xAxis.getRange();
		DateTime minDate = (DateTime) range[0];
		DateTime maxDate = (DateTime) range[1];
		if (!model.getStart().equals(minDate)
				|| !model.getEnd().equals(maxDate)) {
			xAxis.setRange(new DateTime[] { model.getStart(), model.getEnd() },
					false);
			xAxis.setRange(new DateTime[] { model.getStart(), model.getEnd() },
					false);
			chart.refresh();
			rangeFromGraph(true);
		}
	}
        
	public void addNode(CalendarNode node, String title) {
		chart.addNode(node, title);
	}
}
