/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package eu.agrosense.client.calendar.api.internal.gui;

import eu.agrosense.client.calendar.api.event.BarEventHandler;
import eu.agrosense.client.calendar.api.internal.util.ModelInflater;
import eu.agrosense.client.calendar.api.model.CalendarExtras;
import eu.agrosense.client.calendar.api.model.CalendarNode;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.scene.Node;
import javafx.scene.chart.Axis;
import javafx.scene.chart.XYChart;
import javafx.scene.control.Label;
import org.joda.time.DateTime;

public class CalendarChart extends XYChart<DateTime, CalendarNode> {

    private static final DateTime MIN_DATE = new DateTime(0);
    private static final DateTime MAX_DATE = new DateTime(3000, 1, 1, 1, 1); //far far in future
    private static final double MIN_BAR_WIDTH = 20;
    private BarEventHandler barEventHandler;

    public CalendarChart(Axis<DateTime> xAxis, Axis<CalendarNode> yAxis) {
        super(xAxis, yAxis);
        setAnimated(false);
        xAxis.setAnimated(false);
        yAxis.setAnimated(false);

        ObservableList<XYChart.Series<DateTime, CalendarNode>> list = FXCollections.observableArrayList(new ArrayList<XYChart.Series<DateTime, CalendarNode>>());
        setData(list);

        //another hack. we need to remove chart title
        ObservableList<Node> children = getChildren();
        if (!children.isEmpty() && children.get(0) instanceof Label) {
            children.remove(0);
        }
    }

    @Override
    protected void dataItemAdded(Series<DateTime, CalendarNode> series, int itemIndex, Data<DateTime, CalendarNode> item) {

        Node bar = getBar(getData().indexOf(series), item, itemIndex);
        getPlotChildren().add(bar);

        if (series.getNode() != null) {
            series.getNode().toFront();
        }
    }

    @Override
    protected void updateAxisRange() {

        final Axis<DateTime> xAxis = getXAxis();
        final Axis<CalendarNode> yAxis = getYAxis();
        List<DateTime> xData = null;
        List<CalendarNode> yData = null;
        if (xAxis.isAutoRanging()) {
            xData = new ArrayList<>();
        }
        if (yAxis.isAutoRanging()) {
            yData = new ArrayList<>();
        }

        DateTime maxDate = MIN_DATE;
        DateTime minDate = MAX_DATE;

        if (xData != null || yData != null) {
        	
            for (Series<DateTime, CalendarNode> series : getData()) {
                for (Data<DateTime, CalendarNode> data : series.getData()) {

                    CalendarExtras extras = (CalendarExtras) data.getExtraValue();

                    if (xData != null) {
                        if (data.getXValue().isBefore(minDate)) {
                            minDate = data.getXValue();
                        }
                        if (extras.getEndDate().isAfter(maxDate)) {
                            maxDate = extras.getEndDate();
                        }
                    }
                    if (yData != null && data.getYValue().getParent() == null) { //I want only items without parent to be categories 
                        yData.add(data.getYValue());
                    }
                }
            }
            if (xData != null) {
                xData.add(minDate);
                xData.add(maxDate);
                xAxis.invalidateRange(xData);
            }
            if (yData != null) {
                yAxis.invalidateRange(yData);
            }
        }
    }

    @Override
    protected void dataItemChanged(javafx.scene.chart.XYChart.Data<DateTime, CalendarNode> data) {
    	//TODO handle data change
    }

    @Override
    protected void dataItemRemoved(Data<DateTime, CalendarNode> item, Series<DateTime, CalendarNode> series) {
        final Node candle = item.getNode();
        getPlotChildren().remove(candle);
    }

    @Override
    protected void layoutPlotChildren() {
        if (getData() == null) {
            return;
        }

        for (Series<DateTime, CalendarNode> series : getData()) {
            Iterator<Data<DateTime, CalendarNode>> iter = getDisplayedDataIterator(series);

            while (iter.hasNext()) {
                Data<DateTime, CalendarNode> item = iter.next();
                DateTime startValue = getCurrentDisplayedXValue(item);
                double xStart = getXAxis().getDisplayPosition(startValue);
                double yStart = getYAxis().getDisplayPosition(item.getYValue());

                CalendarExtras barExtras = (CalendarExtras) item.getExtraValue();
                DateTime endVale = barExtras.getEndDate();
                double xWidth = getXAxis().getDisplayPosition(endVale) - xStart;

                if (xWidth < MIN_BAR_WIDTH) {
                    xWidth = MIN_BAR_WIDTH;
                }

                Node itemNode = item.getNode();
                if (itemNode instanceof Bar) {
                    Bar bar = (Bar) itemNode;
                    bar.update(xWidth, barExtras.getDepth());
                    bar.updateTooltip(barExtras.getTitle(), barExtras.getCategory(), startValue, endVale);
                    bar.updateLayout(xStart, yStart);
                }
            }
        }
    }

    @Override
    protected void seriesAdded(Series<DateTime, CalendarNode> series, int seriesIndex) {
        for (int i = 0; i < series.getData().size(); i++) {
            Data<DateTime, CalendarNode> item = series.getData().get(i);
            Node bar = getBar(seriesIndex, item, i);
            getPlotChildren().add(bar);
        }
    }

    @Override
    protected void seriesRemoved(Series<DateTime, CalendarNode> series) {
        for (XYChart.Data<DateTime, CalendarNode> d : series.getData()) {
            Node node = d.getNode();
            getPlotChildren().remove(node);
        }
    }

    private Node getBar(int seriesIndex, final Data<DateTime, CalendarNode> item, int itemIndex) {
        Node node = item.getNode();
        CalendarExtras extras = (CalendarExtras) item.getExtraValue();

        if (node instanceof Bar) {
            ((Bar) node).setSeriesAndDataStyleClasses("series" + seriesIndex, "data" + itemIndex);
        } else {

            CalendarNode model = item.getYValue();
            node = new Bar(model, "series" + seriesIndex, "data" + itemIndex, extras.getDepth());
            ((Bar) node).setOnBarClicked(barEventHandler);

            item.setNode(node);
        }
        return node;
    }

    public void refresh() {
        requestChartLayout();
        getXAxis().invalidateRange(null);
        getXAxis().requestAxisLayout();
    }

    public void addNode(CalendarNode node, String title) {
        ModelInflater inflater = new ModelInflater(node);
        XYChart.Series<DateTime, CalendarNode> series = new XYChart.Series<>();
        series.setName(title);
        DateTime[] dates = inflater.getDates();
        CalendarExtras[] extras = inflater.getExtras();
        List<CalendarNode> nodes = inflater.getNodes();

        for (int i = 0; i < nodes.size(); i++) {
            series.getData().add(new XYChart.Data<>(dates[i], nodes.get(i), extras[i]));
        }

        getData().add(series);
    }

    void toFront(Bar bar) {
        bar.setOrder(getPlotChildren().indexOf(bar));
        getPlotChildren().remove(bar);
        getPlotChildren().add(bar);
    }

    void toBack(Bar bar) {
        int oldOrder = bar.getOrder();
        getPlotChildren().remove(bar);
        getPlotChildren().add(oldOrder, bar);
    }

    public void setBarEventHandler(BarEventHandler barEventHandler) {
        this.barEventHandler = barEventHandler;
    }
}
